"""Tests for :mod:`arxiv.util.schema`."""

from unittest import TestCase, mock
import json
import os
import tempfile

from arxiv import util
from arxiv import status


class TestValidateRequest(TestCase):
    """Add request body validation to a Flask route."""

    def setUp(self):
        """Initialize a :class:`.Flask` app and write a schema."""
        _, self.schema_path = tempfile.mkstemp()
        schema_body = json.dumps({
            "title": "FooResource",
            "additionalProperties": False,
            "required": ["baz", "bat"],
            "type": "object",
            "properties": {
                "baz": {
                    "type": "string"
                },
                "bat": {
                    "type": "integer"
                }
            }
        })
        with open(self.schema_path, 'w') as f:
            f.write(schema_body)

    def tearDown(self):
        """Clean up the temporary file used for the schema."""
        os.remove(self.schema_path)

    def test_decorate_route(self):
        """:func:`.schema.validate_request` generates a decorator."""
        decorator = util.schema.validate_request(self.schema_path)
        self.assertTrue(hasattr(decorator, '__call__'))

    @mock.patch('arxiv.util.schema.request')
    def test_validate_with_invalid_request_data(self, mock_request):
        """Decorated route function returns 400 bad request on invalid data."""
        mock_request.get_json = mock.MagicMock(
            return_value={'not': 'what you expected'}
        )

        decorator = util.schema.validate_request(self.schema_path)

        @decorator
        def foo_route():
            return {}, status.HTTP_200_OK, {}

        r_body, r_status, r_headers = foo_route()
        self.assertEqual(r_status, status.HTTP_400_BAD_REQUEST,
                         "Failed to catch an invalid request.")
        self.assertIn('reason', r_body)

    @mock.patch('arxiv.util.schema.request')
    def test_validate_with_valid_request_data(self, mock_request):
        """Decorated route function returns normally on valid data."""
        mock_request.get_json = mock.MagicMock(
            return_value={'baz': 'asdf', 'bat': -1}
        )

        decorator = util.schema.validate_request(self.schema_path)

        @decorator
        def foo_route():
            return {}, status.HTTP_200_OK, {}

        r_body, r_status, r_headers = foo_route()
        self.assertEqual(r_status, status.HTTP_200_OK,
                         "Interfered with a valid request")


class TestValidateSchema(TestCase):
    """Validate data against a JSON Schema specification."""

    def setUp(self):
        """Provision a temporary file for the schema."""
        _, self.schema_path = tempfile.mkstemp()

    def tearDown(self):
        """Clean up the temporary file used for the schema."""
        os.remove(self.schema_path)

    def test_load_valid_schema(self):
        """:func:`.schema.load` returns a callable validator."""
        schema_body = json.dumps({
            "title": "FooResource",
            "additionalProperties": False,
            "required": ["baz", "bat"],
            "type": "object",
            "properties": {
                "baz": {
                    "type": "string"
                },
                "bat": {
                    "type": "integer"
                }
            }
        })

        with open(self.schema_path, 'w') as f:
            f.write(schema_body)

        try:
            validator = util.schema.load(self.schema_path)
        except Exception as e:
            self.fail('Failed to load valid schema: %s' % e)
        self.assertTrue(hasattr(validator, '__call__'))

        try:
            validator({'baz': 'asdf', 'bat': -1})
        except util.schema.ValidationError as e:
            self.fail('Validation failed on valid data: %s' % e)

    def test_load_invalid_schema(self):
        """:func:`.util.schema.load` raises IOError if schema is not valid JSON."""
        with open(self.schema_path, 'w') as f:
            f.write('thisisnotajsons..."."{{chema')

        with self.assertRaises(IOError):
            util.schema.load(self.schema_path)

    def test_load_nonexistant_schema(self):
        """:func:`.util.schema.load` raises IOError if schema does not exist."""
        with self.assertRaises(IOError):
            util.schema.load('/tmp/is/unlikely/to/exist')

    def test_validation_failed(self):
        """Validator raises :class:`.util.schema.ValidationError` on bad data."""
        schema_body = json.dumps({
            "title": "FooResource",
            "additionalProperties": False,
            "required": ["baz", "bat"],
            "type": "object",
            "properties": {
                "baz": {
                    "type": "string"
                },
                "bat": {
                    "type": "integer"
                }
            }
        })
        with open(self.schema_path, 'w') as f:
            f.write(schema_body)

        validator = util.schema.load(self.schema_path)

        with self.assertRaises(util.schema.ValidationError):
            validator({'nobody': 'expects', 'thespanish': 'inquisition'})



"""
Support for typed flash messages.

Flask provides a `simple cookie-based flashing mechanism
<http://flask.pocoo.org/docs/1.0/patterns/flashing/>`. This module extends
that mechanism to support structured messages (i.e. dicts) and set
message categories/severity.

.. note::

   For security purposes, messages are not treated as safe in the template by
   default. To treat a message as safe, use ``safe=True`` when generating the
   flash notification.


For example:

.. code-block:: python

   from flask import url_for
   from arxiv.base import messages

   def handle_request():
       ...

       help_url = url_for('help')
       messages.flash_warning(f'This is a warning, see <a href="{help_url}">'
                              f'the docs</a> for more information',
                              title='Warning title', safe=True)
       messages.flash_info('This is some info', title='Info title')
       messages.flash_failure('This is a failure', title='Failure title')
       messages.flash_success('This is a success', title='Success title')
       messages.flash_warning('This is a warning that cannot be dismissed',
                              dismissable=False)

"""

from typing import Optional
from flask import flash, Markup

INFO = 'info'
WARNING = 'warning'
FAILURE = 'danger'   # This is odd, but we use `danger` in styles.
SUCCESS = 'success'


def _flash_with(severity: str, message: str, title: Optional[str] = None,
                dismissable: bool = True, safe: bool = False) -> None:
    if safe:
        message = Markup(message)
    data = {'message': message, 'title': title, 'dismissable': dismissable}
    flash(data, severity)


def flash_info(message: str, title: Optional[str] = None,
               dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash an informative message to the user.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(INFO, message, title, dismissable, safe)


def flash_warning(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a warning message to the user.

    Warnings are like info messages, but with a higher level of concern. For
    example, this might be used to notify a user that an unintended consequence
    might be imminent.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(WARNING, message, title, dismissable, safe)


def flash_failure(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a failure message to the user.

    Failures are an unavoidable part of life. We should be honest about our
    mistakes, and find constructive ways to move forward with our lives. That
    means that we should provide some specific hints for how the user can
    achieve success (nb: this is an important accessibility requirement).

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(FAILURE, message, title, dismissable, safe)


def flash_success(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a success message to the user.

    It's nice to know when you've done something right for a change.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(SUCCESS, message, title, dismissable, safe)



"""
Provides routes for verifying base templates and exception handling.

The blueprint instantiated here is **not** for use in a production application;
it is not attached by :class:`arxiv.base.Base`.
"""

from typing import Any, Tuple, Callable, Dict
from flask import Blueprint, render_template, current_app, make_response, \
    Response, flash, url_for

from arxiv import status
from arxiv.base.exceptions import NotFound, Forbidden, Unauthorized, \
    MethodNotAllowed, RequestEntityTooLarge, BadRequest, InternalServerError

from . import messages

blueprint = Blueprint('ui', __name__, url_prefix='')


@blueprint.route('/styleguide', methods=['GET'])
def test_page() -> Response:
    """Render the test page."""
    rendered = render_template("base/styleguide.html", pagetitle='Home')
    response = make_response(rendered, status.HTTP_200_OK)

    # Demonstrate flash messages. To see these messages, reload the page.
    help_url = url_for('help')
    messages.flash_warning(f'This is a warning, see <a href="{help_url}">the'
                           f' docs</a> for more information',
                           title='Warning title', safe=True)
    messages.flash_info('This is some info', title='Info title')
    messages.flash_failure('This is a failure', title='Failure title')
    messages.flash_success('This is a success', title='Success title')
    messages.flash_warning('This is a warning that cannot be dismissed',
                           dismissable=False)
    return response


@blueprint.route('/404', methods=['GET'])
def test_404() -> Response:
    """Test the 404 error page."""
    raise NotFound()


@blueprint.route('/403', methods=['GET'])
def test_403() -> Response:
    """Test the 403 error page."""
    raise Forbidden()


@blueprint.route('/401', methods=['GET'])
def test_401() -> Response:
    """Test the 401 error page."""
    raise Unauthorized()


@blueprint.route('/405', methods=['GET'])
def test_405() -> Response:
    """Test the 405 error page."""
    raise MethodNotAllowed()


@blueprint.route('/413', methods=['GET'])
def test_413() -> Response:
    """Test the 413 error page."""
    raise RequestEntityTooLarge()


@blueprint.route('/400', methods=['GET'])
def test_400() -> Response:
    """Test the 400 error page."""
    raise BadRequest()


@blueprint.route('/500', methods=['GET'])
def test_500() -> Response:
    """Test the 500 error page."""
    raise InternalServerError()
